// 
// QClam
// Copyright(C) Sergey A. Galin, 2004-2007
// http://sageshome.net/oss/
// Please see COPYING.txt for legal information.
//

#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <unistd.h>
#include <sys/stat.h>
#include <time.h>
#include <sys/timeb.h>
#include <errno.h>

// Uncomment to have debug messages printed
//#define QCLAM_DEBUG

#define QCLAM_VERSION "0.8"
#define QCLAM_BUFFER_SIZE (1024*256)
#define QCLAM_DEFAULT_CLAMSCAN "/usr/bin/clamscan"
#define QCLAM_DEFAULT_LOGNAME "qclam.log"

#define QCLAM_CLAMSCAN_DETECT "clamscan"
#define QCLAM_CLAMDSCAN_DETECT "clamdscan"
// #define QCLAM_CLAMSCAN_OPTIONS "--stdout --mbox"
#define QCLAM_CLAMSCAN_OPTIONS "--stdout"
#define QCLAM_CLAMDSCAN_OPTIONS "--stdout"

#define MAX_PATH 256
#define MAX_COMMAND (MAX_PATH*3)
#define MAX_VIRUS_INFO 1024

#define DEBUG(x) {if(Debug) printf("DEBUG: %s\n", x); }

#define USEMKSTEMP

#if !defined(USEMKSTEMP)
  // Allow --group option
  #define ENABLE_GROUP
#endif  


/*
       "man qmail_command" reads:
       
       SENDER  is  the envelope sender address.  NEWSENDER is the
       forwarding envelope sender address, as described  in  dot-
       qmail(5).   RECIPIENT  is  the envelope recipient address,
       local@domain.  USER is user.  HOME is your home directory,
       homedir.   HOST  is  the  domain  part  of  the  recipient
       address.  LOCAL is the local part.   EXT  is  the  address
       extension, ext.

       HOST2 is the portion of HOST preceding the last dot; HOST3
       is the portion of HOST preceding the  second-to-last  dot;
       HOST4  is  the portion of HOST preceding the third-to-last
       dot.

       EXT2 is the portion of EXT following the first dash;  EXT3
       is the portion following the second dash; EXT4 is the por-
       tion following the third dash.   DEFAULT  is  the  portion
       corresponding  to the default part of the .qmail-...  file
       name; DEFAULT is not set if the file  name  does  not  end
       with default.

       DTLINE  and  RPLINE are the usual Delivered-To and Return-
       Path lines, including newlines.  UFLINE is the  UUCP-style
       From_ line that qmail-local adds to mbox-format files.
*/

const int QMAIL_DONTPROCESS=99;
const int CLAMAV_VIRUS=1;
int Verbose=0, Debug=0;




char *trim(char *s){
  int i;
  for(i=strlen(s); i>=0 && s[i]<32; i--) 
    s[i]=0;
  return s;
}




char *nextline(char *s, char *outbuf){
  while(*s){
    if(*s=='\n'){ // LF found, end copying and return current position in *s
      *outbuf=0;
      return s+1;
    }
    // not LF or '\0' => copy character from s to outbuf
    *outbuf++=*s++;
  }
  *outbuf=0;
  return NULL;
}




const char *clamoptions(const char *program){
  if(strstr(program, QCLAM_CLAMDSCAN_DETECT)!=NULL)
    return QCLAM_CLAMDSCAN_OPTIONS;
  else    
    return QCLAM_CLAMSCAN_OPTIONS;
}




int fileexists(const char *fn){
  struct stat st;
  if(stat(fn, &st)) //if(stat(fn, &st)==ENOENT) 
    return 0;
  else
    return 1;
}




void infectedname(const char *dir, char *out){
  time_t t;
  struct tm* ts;
  int count=0;
  time(&t);
  ts=localtime(&t);
  do{
    sprintf(out, "%s/INFECTED-%04d%02d%02d-%02d%02d%02d-%03d.mbox", dir, 
      ts->tm_year+1900, 
      ts->tm_mon+1,
      ts->tm_mday,
      ts->tm_hour,
      ts->tm_min,
      ts->tm_sec,
      count);
    count++;
    DEBUG(out);
  }while(fileexists(out));  
}



int main(int argc, char *argv[]){
  char *buf, *homedir, *sender, *recipient;
  size_t i, bufsz=0, sz=QCLAM_BUFFER_SIZE;
  
  #if defined(USEMKSTEMP)
    int temp_fd;
  #endif
      
  buf=(char*)malloc(sz);
  if(buf==NULL){
    fprintf(stderr, "Failed to allocate initial buffer!\n");
    return 0;
  }
  
  int 
    chr, 
    exitcode=0, 
    avcode=0, 
    tmperror=0, 
    av_run=0, 
    log_stderr=0, 
    av_error=0, 
    keep_tmp=0, 
    copy_infected=0,
    copy_error=0,
    #if defined(ENABLE_GROUP)
      chgroup=-1,
    #endif  
    indatasz, 
    indataptr,
    in_line_length;

  char 
    filename[MAX_PATH*2]="", 
    infecteddir[MAX_PATH]="",
    infectedfile[MAX_PATH*2]="",
    tmpdir[MAX_PATH]="/tmp",
    clamscan[MAX_PATH]=QCLAM_DEFAULT_CLAMSCAN,
    command[MAX_COMMAND]="", 
    logfile[MAX_PATH]="", 
    virinfo[MAX_VIRUS_INFO]="",
    *indata,
    *inptr,
    in_line[MAX_COMMAND];
    
    
  FILE *tmp, *log, *copy, *process;
  
  // Parse command line arguments
  for(i=1; i<argc; i++){
    if(!strcmp(argv[i], "--help")){
      printf(
        "QClam " QCLAM_VERSION ", build date: " __DATE__ ", " __TIME__ "\n"
	"Copyright (C) Sergey A. Galin, 2004-2007; Homepage: http://sageshome.net/\n"
        "This is a free software; see COPYING.txt for copying conditions.  There is NO\n"
        "warranty; not even for MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.\n"
	"Project Homepage: http://sageshome.net/oss/qclam.php\n"
	"\n"
	"This program is intented to plug ClamAV antivirus into QMail amd scan messages\n"
	"coming into your mailbox for viruses. Add the following two lines into your\n"
	"~/.qmail file:\n"
	"\n"
	"|/usr/local/bin/qclam\n"
        "./Maildir/\n"
	"\n"
	"Please refer to QClam README.txt file and QMail documentation for more info.\n"
	"\n"
	"Optional command-line arguments:\n"
	"\n"
	"  --help                     - print this information and exit.\n"
	"  --version                  - print QClam version number only and exit.\n"
	"  --clam <path-to-clamscan>  - use another clamscan or clamdscan execultable;\n"
	"                               default: \"" QCLAM_DEFAULT_CLAMSCAN "\".\n"
	"  --temp <temporary-dir>     - directory for temorary files.\n"	
	"  --infected <dir>           - copy infected messages into the specified directory.\n"	
	"  --log <log-file>           - specify QClam log file name.\n"
	"                               default: \"~/" QCLAM_DEFAULT_LOGNAME "\".\n" 
	"\n"
	"Debug options:\n"
	"\n"	
	"  --log_stderr               - copy lines to be written into log to stderr.\n"		
// !! No verbose messages yet, so don't show -v in help :)	
//	"  -v                         - verbose output\n"
	"  --debug                    - debug mode.\n"
	"  --keep_tmp                 - don't remove temporary files.\n"
	"\n"	
	
        #if defined(ENABLE_GROUP)	
  	  "Experimental options:\n"
	  "\n"	
	  "  --group <numeric-gid>      - change effective process group and allows the group\n"
	  "                               to read temorary file (useful for clamdscan).\n"
	  "\n"
	#endif  
      );
      return 0;
      
    }else if(!strcmp(argv[i], "--version")){      
      puts(QCLAM_VERSION);
      return 0;
      
    }else if(!strcmp(argv[i], "--clam")){      
      if(i==argc-1){
        fprintf(stderr, "--clam option specified, but no path to ClamAV executable.\n");
	return 0;
      }
      strncpy(clamscan, argv[++i], sizeof(clamscan)-1);      

    }else if(!strcmp(argv[i], "--temp")){      
      if(i==argc-1){
        fprintf(stderr, "--temp option specified, but no temp directory name.");
	return 0;
      }
      strncpy(tmpdir, argv[++i], sizeof(tmpdir)-1);      

    }else if(!strcmp(argv[i], "--infected")){      
      if(i==argc-1){
        fprintf(stderr, "--infected option specified, but no temp directory name.");
	return 0;
      }
      strncpy(infecteddir, argv[++i], sizeof(infecteddir)-1);      
      copy_infected=1;
            
    #if defined(ENABLE_GROUP)  
      }else if(!strcmp(argv[i], "--group")){      
        if(i==argc-1){
          fprintf(stderr, "--group option specified, but no group name.\n");
      	  return 0;
        }
        chgroup=0;
        if(sscanf(argv[++i], "%d", &chgroup)!=1){
          fprintf(stderr, "--group option specified, group ID is not an integer value.\n");
	  return 0;
        }
    #endif	
      
    }else if(!strcmp(argv[i], "--log")){      
      if(i==argc-1){
        fprintf(stderr, "--log option specified, but no path to log file.\n");
	return 0;
      }
      strncpy(logfile, argv[++i], sizeof(logfile)-1);            
      
    }else if(!strcmp(argv[i], "--log_stderr")){      
      log_stderr=1;
      
    }else if(!strcmp(argv[i], "-v")){            
      Verbose++;
      
    }else if(!strcmp(argv[i], "--debug")){            
      Debug++;     

    }else if(!strcmp(argv[i], "--keep_tmp")){            
      keep_tmp=1;     
      
    }else{
      fprintf(stderr, "Unknown command-line argument: %s. Try \"qclam --help\" for information.\n", argv[i]);
      return 0;
    }
  }
  
  #if defined(ENABLE_GROUP)
    if(chgroup>=0){
      if(setgid(chgroup)){
        fprintf(stderr, "Failed to set process group ID to %d.\n", chgroup);
        return 0;
      };
    }
  #endif    
  
  // Get message data from stdin.
  // sz is the allocated buffer size, bufsz is the used buffer size.
  while(!feof(stdin)){
    if(bufsz>=sz-1){ // We almost filled the buffer, need to get more memory...
      buf=(char*)realloc(buf, sz+=1024*256);
      if(buf==NULL){
        fprintf(stderr, "Failed to allocate additional buffer!\n");
        return 0;
      }
    }
    chr=fgetc(stdin);
    if(chr==EOF) break; else buf[bufsz++]=chr;
  }
  buf[bufsz]=0;

  sender=getenv("SENDER"); if(sender==NULL) sender="Unknown-sender";
  recipient=getenv("RECIPIENT"); if(recipient==NULL) recipient="Unknown-recipient";

  // Saving message into temporary file
  // sprintf(filename, "/tmp/qclam_temp_%u", getpid());
  // Stupid man doesn't have any recommendation how to free memory from
  // tempnam().
  #if !defined(USEMKSTEMP)
    strncpy(filename, tempnam(tmpdir, "qclam"), sizeof(filename)-1);  
    tmp=fopen(filename, "w");
    
  #else    
    sprintf(filename, "%s/qclamXXXXXX", tmpdir);
    temp_fd = mkstemp(filename);
    DEBUG(filename);    
    if(temp_fd == -1){
      tmp=NULL;
    }else{
      tmp=fdopen(temp_fd, "w+");
    }
  #endif
      

  if(tmp!=NULL){
    #if defined(ENABLE_GROUP)
      if(chgroup>=0){
        chmod(filename, 0640);
      }else{
        chmod(filename, 0600);
      }  
    #else
      chmod(filename, 0600);
    #endif  
      
    fwrite(buf, bufsz, 1, tmp);
    fclose(tmp);

    sprintf(command, "%s %s %s", clamscan, clamoptions(clamscan), filename);
    //avcode=system(command);

    // Scan temp file using antivirus and get its output
    process=popen(command, "r");
    if(process==NULL)
      avcode=-1;
    else{
      indatasz=1024*4; 
      indataptr=0;
      indata=(char*)malloc(indatasz);
      if(indata==NULL){ 
        fprintf(stderr, "Failed to allocate process input buffer!\n"); 
	if(!keep_tmp) 
	  unlink(filename); 
	return 0; 
      }
      while(!feof(process)){
        // Making sure we have bytes for data+trailing zero
        if(indataptr>=indatasz-1){
          indata=(char*)realloc(indata, indatasz+=1024*256);
          if(indata==NULL){
            fprintf(stderr, "Failed to allocate additional process input buffer!\n");
            if(!keep_tmp)
	      unlink(filename);
            return 0;
          }
        }
        chr=fgetc(process);
        if(chr==EOF) break; else indata[indataptr++]=chr;
      }
      // Adding trailing zero at end of buffer
      indata[indatasz]=0;
      pclose(process);
      //printf("GOT OUT:\n%s\n", indata);
      // Scanning antivirus output...
      inptr=indata;
      
      for(;;){
        inptr=nextline(inptr, in_line);
	if(inptr==NULL) break;
	DEBUG(in_line); 	
        trim(in_line);
        //if(*in_line==0) break;	
	if((in_line_length=strlen(in_line))<8) continue;	
        // FOUND must be end of the line
        if(!strcmp(in_line+in_line_length-6, " FOUND")){
          if(virinfo[0]) strcat(virinfo, "; ");
          strcat(virinfo, in_line);
          avcode=CLAMAV_VIRUS;
	  av_run=1;
	  DEBUG("VIRUS FOUND!"); 
        }else if(!strcmp(in_line+in_line_length-6, " ERROR")){
	  av_run=1;
	  av_error=1;
	  DEBUG("ERROR FOUND!");
        }else if(!strcmp(in_line+in_line_length-3, " OK")){
	  av_run=1;
	  DEBUG("OK FOUND");
        }
      }
      free(indata);
    }
    // If AV didn't print "SCAN SUMMARY", it was not AV we need.
    if(!av_run){ 
      avcode=-1;
      DEBUG("AV RUN NOT DETECTED!"); 
    }  
    if(avcode==CLAMAV_VIRUS){
      exitcode=QMAIL_DONTPROCESS;
      DEBUG("avcode==CLAMAV_VIRUS, setting exitcode accordingly");
      
      // Writing a copy of the infected message
      if(copy_infected){
        //strncpy(infectedfile, tempnam(infecteddir, "virus"), sizeof(filename)-1);
	DEBUG("MAKING A COPY OF THE MESSAGE");
	infectedname(infecteddir, infectedfile);
        copy=fopen(infectedfile, "w");
        if(copy!=NULL){
	  #if defined(ENABLE_GROUP)
            if(chgroup>=0){
              chmod(filename, 0640);
            }else{
              chmod(filename, 0600);
            }  
          #else	    
	    chmod(filename, 0600);
	  #endif    
          fwrite(buf, bufsz, 1, copy);
          fclose(copy);
	}else
	  copy_error=1;  
	DEBUG("...DONE");  
      }
      
    }  
    if(!keep_tmp)
      unlink(filename);
  }else{
    tmperror=1;
  }

  // Write to log file...
  homedir=getenv("HOME");
  if(homedir!=NULL || logfile[0]){
    time_t t;
    time(&t);
    char timestr[64];
    ctime_r(&t, timestr);
    trim(timestr);
    
    // If no log file specified, write into home directory
    if(!logfile[0])
      sprintf(logfile, "%s/%s", homedir, QCLAM_DEFAULT_LOGNAME);
    
    log=fopen(logfile, "a");
    if(log!=NULL){
      if(exitcode==0){
        
        // No virus found or AV error.
	// (We have to return 0 in case of AV errors, or all e-mail messages will be lost!)
        if(avcode<0){
          fprintf(log, "%s AV_FAILED: From %s to %s, %d byte(s)\n", timestr, sender, recipient, bufsz);
	  if(log_stderr) fprintf(stderr, "%s AV_FAILED: From %s to %s, %d byte(s)\n", timestr, sender, recipient, bufsz);
        }else if(tmperror){
          fprintf(log, "%s TEMP_ERROR: From %s to %s, %d byte(s)\n", timestr, sender, recipient, bufsz);
	  if(log_stderr) fprintf(stderr, "%s TEMP_ERROR: From %s to %s, %d byte(s)\n", timestr, sender, recipient, bufsz);
	}else if(av_error){  
          fprintf(log, "%s AV_ERROR: From %s to %s, %d byte(s)\n", timestr, sender, recipient, bufsz);
	  if(log_stderr) fprintf(stderr, "%s AV_ERROR: From %s to %s, %d byte(s)\n", timestr, sender, recipient, bufsz);	
        }else{
          fprintf(log, "%s OK: From %s to %s, %d byte(s)\n", timestr, sender, recipient, bufsz);
	  if(log_stderr) fprintf(stderr, "%s OK: From %s to %s, %d byte(s)\n", timestr, sender, recipient, bufsz);
	};
      }else{
      
        // Virus found.
        fprintf(log, "%s VIRUS: From %s to %s, %d byte(s): %s\n", timestr, sender, recipient, bufsz, virinfo);
	if(log_stderr) fprintf(stderr, "%s VIRUS: From %s to %s, %d byte(s): %s\n", timestr, sender, recipient, bufsz, virinfo);

        if(copy_infected){
  	  if(copy_error){
            fprintf(log, "%s WARNING: Failed to save a copy of infected message to: %s\n", timestr, infectedfile);
	    if(log_stderr) fprintf(stderr, "%s WARNING: Failed to save a copy of infected message to: %s\n", timestr, infectedfile);
          }else{	  
            fprintf(log, "%s SAVE_INFECTED: %s\n", timestr, infectedfile);
  	    if(log_stderr) fprintf(stderr, "%s SAVE_INFECTED: %s\n", timestr, infectedfile);
	  };    
	};  
	
      }	
      fclose(log);
    }else
      fprintf(stderr, "Failed to open log file!\n");
  }else{
    fprintf(stderr, "Cannot find user's home directory!\n");
  }

  free(buf);
  return exitcode;
}

